/*********************************************************************
This is an example for our Monochrome OLEDs based on SSD1306 drivers

  Pick one up today in the adafruit shop!
  ------> http://www.adafruit.com/category/63_98

This example is for a 128x64 size display using I2C to communicate
3 pins are required to interface (2 I2C and one reset)

Adafruit invests time and resources providing this open source code, 
please support Adafruit and open-source hardware by purchasing 
products from Adafruit!

Written by Limor Fried/Ladyada  for Adafruit Industries.  
BSD license, check license.txt for more information
All text above, and the splash screen must be included in any redistribution

https://github.com/rogerclarkmelbourne/Arduino_STM32/tree/master/STM32F1/libraries/Adafruit_SSD1306
*********************************************************************/
#define REVISION 16

#include <SPI.h>
#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306_STM32.h>
#include <curveFitting.h>

#define OLED_RESET 4
Adafruit_SSD1306 display(OLED_RESET);
#define NUMFLAKES 10
#define XPOS 0
#define YPOS 1
#define DELTAY 2


#define LOGO16_GLCD_HEIGHT 16 
#define LOGO16_GLCD_WIDTH  16 
static const unsigned char PROGMEM logo16_glcd_bmp[] =
{ B00000000, B11000000,
  B00000001, B11000000,
  B00000001, B11000000,
  B00000011, B11100000,
  B11110011, B11100000,
  B11111110, B11111000,
  B01111110, B11111111,
  B00110011, B10011111,
  B00011111, B11111100,
  B00001101, B01110000,
  B00011011, B10100000,
  B00111111, B11100000,
  B00111111, B11110000,
  B01111100, B11110000,
  B01110000, B01110000,
  B00000000, B00110000 };

#if (SSD1306_LCDHEIGHT != 64)
#error("Height incorrect, please fix Adafruit_SSD1306.h!");
#endif

#include <JC_Button.h>// https://github.com/JChristensen/JC_Button
Button pulse_button(PB12, 350);// 350ms de-bounce
int LONG_PRESS = 1000;// Button press 1 sec

// User select variables
float sat_max = 40;// percent
float min_IL = .32;// Minimum current threshold for establishing initial inductance with acceptable error. See Table below
 /* Inductor(A)| +/-Error |
 *    0.16    |   10%    |
 *    0.32    |   5%     |
 *    0.81    |   2%     |
 *    1.62    |   1%     |
 */
float RL = 0.0;// Inductor resistance
const float R_sns = .05;// current sense resistor value(Ω)
float sat_0 = 5;// Inductor saturation value(%) used to calculate the slope of the linear line
float sat_1 = 10;// Saturation threshold value(%)
float sat_2 = 20;// Saturation threshold value(%)
float sat_3 = 30;// Saturation threshold value(%)
int poly_samples = 32;//
const int order = 3; 

int pulse_period = 10; //pulse duration in microseconds
float base_cnt = 0.0;

static unsigned int analog_data[180];// analogRead current data array max pulse_period of 900 / 5us per measurement
static unsigned int interval[180];// Timer1 count for each inductor current reading starting from 0
static float current_data[180];// Inductor current in Amps data array
static float time_data[180];// The time in microseconds for each inductor current reading
static float L_data[180];// Inductor value in microhenrys data array
static float sat_data[180];// Inductor saturation in % data array
static float VL_data[180];// Inductor saturation in % data array

float Vin = 0.0;
float V_open_ckt = 0.0;

int samples;// number of actual ADC reads
float L = 0.0;// Measured Inductance values in microhenrys

// Used in calculating the initial starting inductance
float initial_L = 0.0;// Starting inductance value in microhenries at current >= min_IL
float initial_I = 0.0;// Starting Inductor current Amps at current >= min_IL
float initial_t = 0.0;// Starting time microseconds at current >= min_IL

// Used in calculating the saturation percentage values
float sat_val_0;
float sat_val_1;
float sat_val_2;
float sat_val_3;
float sat_val_last = 0.0;
float I_sat_val_0 = 0.0;
float t_sat_val_0 = 0.0;
float I_sat_val_1 = 0.0;
float t_sat_val_1 = 0.0;
float I_sat_val_2 = 0.0;
float t_sat_val_2 = 0.0;
float I_sat_val_3 = 0.0;
float t_sat_val_3 = 0.0;
float VL;

// For polynomial calculations
double t_poly[64];
double IL_poly[64];
double coeffs[(order + 1)];
double a;
double b;
double c;
double d;
double e;
double f;
double g;

// For parsing the input from the console
const byte numChars = 32;
char receivedChars[numChars];
char tempChars[numChars];// temporary array for use when parsing
// variables to hold the parsed data
char messageFromPC[numChars] = {0};
int integerFromPC = 0;
float float1FromPC = 0.0;
float float2FromPC = 0.0;
char* line_on_off = "on";
char* cf_on_off = "on";// curve fit
bool newData = false;


bool linear_on = true;
bool poly_on = true;
bool serial_data = true;
bool ok;
bool inductance_only;
bool button_long_press;
bool no_inductor;
bool Vin_ok;
bool V_chk;
bool I_chk;
bool found_0 = false;
bool found_1 = false; 
bool found_2 = false; 
bool found_3 = false; 

void setup() {
    Serial.begin(115200);
  
    display.begin(SSD1306_SWITCHCAPVCC, 0x3C);  // initialize with the I2C addr 0x3D (for the 128x64)
    // display the Adafruit logo
    display.display();
    delay(2000);

    // Clear the buffer.
    display.clearDisplay();
    display.display(); 
    // Character size (8,8) (row,col) 8 rows & 16 columns upper left (0,0) Lower right (120,56)
    display.setTextSize(1);
    display.setTextColor(WHITE);
    display.setCursor(0,0);
    display.println("     Inductance"); 
    display.setCursor(0,8);
    display.println("  Saturation Tester"); 
    display.setCursor(0,16);
    display.println("        Rev"); 
    display.setCursor(68,16);
    display.println(REVISION);
    display.setCursor(0,8); 
    display.setCursor(0,32);
    display.println("SHORT Press - L Only");       
    display.setCursor(0,48);
    display.println("LONG Press - SAT Data"); 
    display.display(); 
    delay(1000);

     
    pinMode(PB9, OUTPUT);// Pulse Output
    gpio_write_bit(GPIOB,9,0);//Output pulse LOW    
    pinMode(PB12, INPUT_PULLUP);// Pulse Button
    pinMode(PA0, INPUT_ANALOG);
    pinMode(PA1, INPUT_ANALOG);

    // Clock freq = 72,011,955
    Timer1.setPrescaleFactor(1);
    Timer1.setOverflow(65535);// 1.0988kHz

    adc_set_sample_rate(ADC1, ADC_SMPR_1_5);// 2.5us read time
}

  void loop() {
    
    pulse_button.read();
    
    if (pulse_button.wasReleased())
    {
      // Check for under voltage and verify an inductor for test is connected
      status_check();
      if(!button_long_press)
      {
      inductance_only = true;
      display.clearDisplay();
      display.display();
      display.setCursor(36,0);
      display.println("PROCESSING");
      display.setCursor(22,16);
      display.println("INDUCTANCE ONLY");
      display.display();
      delay(1000);             
      pulse_period = 10;
      console();
      display.clearDisplay();
      display.display();      
      disply();
      }
      else
        button_long_press = false;
    }
    if (pulse_button.pressedFor(LONG_PRESS))
    { 
      // Check for under voltage and verify an inductor for test is connected
      status_check(); 
      inductance_only = false;
      button_long_press = true;
      display.clearDisplay();
      display.display();
      display.setCursor(36,0);
      display.println("PROCESSING");
      display.setCursor(22,16);
      display.println("SATURATION TEST");
      display.display();
      delay(1000);             
      pulse_period = 10;
      console();
      display.clearDisplay();
      display.display();      
      disply();
    }
  }
    void status_check(){
      // Check that proper power supply voltage is present
      // Check to see if the inductor under test is connected ( RL  < 5kΩ)
      // chk flags allow for the display to indicate status one time per event and not flash continuously
      
      while(Vin <= 11.9 || V_open_ckt < .75)
      {
        Vin = analogRead(PA1)* .00082 * 6.688;
        V_open_ckt = analogRead(PA2)* .00082;
        //12V regulator output test ensures external 15-24V power supply is connected and turned on   
        if(Vin <= 11.9)
        {
          if(!V_chk)
          {
            display.clearDisplay();
            display.display();
            display.setCursor(24,0);
            display.println("UNDER VOLTAGE");
            display.setCursor(42,16);
            display.println(Vin);
            display.setCursor(72,16);
            display.println("V");            
            display.display();
            V_chk = true; 
            I_chk = false; 
          } 
        }
        else if(V_open_ckt < .75)
        {
          if(!I_chk)
          {
            display.clearDisplay();
            display.display();
            display.setCursor(0,16);
            display.println("NO INDUCTOR DETECTED");
            display.display();
            I_chk = true;
            V_chk = false;   
          }        
        }
        delay(100);
      }
      Vin = 0;
      V_open_ckt = 3;
      V_chk = false;
      I_chk = false;
    }
    
    void console(){
    if(Serial)
    {
      // OLED Display
      display.setCursor(5,32);
      display.println("SERIAL MONITOR OPEN");
      display.setCursor(15,48);
      display.println("USE THE CONSOLE");
      display.display();
      
      if(inductance_only)
      {      
      Serial.println("");
      Serial.println("");
      Serial.println("");
      Serial.println("");      
      Serial.println("                                         Welcome to the Console Menu");
      delay(5);
      Serial.println("                                             Inductance Only Test");
      Serial.println("");
      delay(5);
      Serial.println("Simply copy, paste and edit (include the delimiter <>) and press send.  NOTE: The current value is shown");
      delay(5);
      Serial.println("When finished send <ok> to continue.");
      delay(5);
      Serial.println("");
      delay(5);
      Serial.print("1) <RL,");
      delay(5);
      Serial.print(RL,3);
      delay(5);
      Serial.println("> - In order to get a more accurate reading, measure or estimate the inductors resistance(RL).");
      delay(5);
      Serial.print("2) <Minimum Inductor current,");
      Serial.print(min_IL);
      Serial.println("> - Establish the IL threshold for the initial inductance calculation");
      delay(5);

      
      }
      else
      {
      Serial.println("");
      Serial.println("");
      Serial.println("");
      Serial.println("");
      Serial.println("                                         Welcome to the Console Menu");
      delay(5);
      Serial.println("                                         Inductance Saturation Test");
      delay(5); 
      Serial.println("");     
      Serial.println("Here you can choose various options and send them one by one.");
      delay(5);
      Serial.println("Simply copy, paste and edit (include the delimiter <>) and press send. NOTE: The current value is shown.");
      delay(5);
      Serial.println("To turn off a feature change from on to off");
      delay(5);
      Serial.println("When finished send <ok> to continue.");
      delay(5);
      Serial.println("");
      delay(5);
      Serial.print("1) <RL,");
      delay(5);
      Serial.print(RL,3);
      delay(5);
      Serial.println("> - In order to get a more accurate reading, measure or estimate the inductors resistance(RL).");
      delay(5);
      Serial.print("2) <Minimum Inductor current,");
      delay(5);
      Serial.print(min_IL);
      delay(5);
      Serial.println("> - Establish the IL threshold for the initial inductance calculation");
      delay(5);
      Serial.print("3) <linear_line_");
      delay(5);
      Serial.print(line_on_off);
      delay(5);
      Serial.print(",");
      delay(5);            
      Serial.print(sat_0);
      delay(5);
      Serial.println("> - plots a straight line from zero through the desired saturation percentage point");
      delay(5);
      Serial.print("4) <curve_fit_");
      delay(5);      
      Serial.print(cf_on_off);
      delay(5);
      Serial.print(",");
      delay(5);
      Serial.print(poly_samples);
      delay(5);
      Serial.println("> - Performs a 3rd order polynomial curve fit when samples are less than the desired # of samples [64 max]");
      Serial.print("5) <saturation_limit,");
      delay(5); 
      Serial.print(sat_max);
      delay(5);
      Serial.println("> - Establish a saturation % limit when either reached or exceeded will terminate the pulse increment");
      delay(5);
      Serial.print("6) <First_Saturation_detect,");
      delay(5); 
      Serial.print(sat_1);
      delay(5);
      Serial.println("> - The first saturation % level when reached will be displayed along with the corresponding IL value");
      delay(5);
      Serial.print("7) <Second_Saturation_detect,");
      delay(5); 
      Serial.print(sat_2);
      delay(5);
      Serial.println("> - The second saturation % level when reached will be displayed along with the corresponding IL value");
      delay(5);
      Serial.print("8) <Third_Saturation_detect,");
      delay(5); 
      Serial.print(sat_3);
      delay(5);
      Serial.println("> - The third saturation % level when reached will be displayed along with the corresponding IL value");
      delay(5);
      Serial.println("");           
      delay(5);
      }
     while(!ok)
     {
      recvWithStartEndMarkers();
      if (newData == true)
      {
        strcpy(tempChars, receivedChars);
        // this temporary copy is necessary to protect the original data
        // because strtok() used in parseData() replaces the commas with \0
        parseData();
        apply_settings();
        newData = false;
      }

     }
    }
    pulse();
  }
    void recvWithStartEndMarkers() 
    {
      static boolean recvInProgress = false;
      static byte ndx = 0;
      char startMarker = '<';
      char endMarker = '>';
      char rc;
      while (Serial.available() > 0 && newData == false) 
      {
        rc = Serial.read();
        if (recvInProgress == true) {
            if (rc != endMarker) {
                receivedChars[ndx] = rc;
                ndx++;
                if (ndx >= numChars) {
                    ndx = numChars - 1;
                }
            }
            else {
                receivedChars[ndx] = '\0'; // terminate the string
                recvInProgress = false;
                ndx = 0;
                newData = true;
            }
        }
        else if (rc == startMarker) {
            recvInProgress = true;
        }
      }
    }

    void parseData()
    // split the data into its parts
    {      
      char * strtokIndx;                   // this is used by strtok() as an index
      strtokIndx = strtok(tempChars,",");  // get the first part - the string
      strcpy(messageFromPC, strtokIndx);   // copy it to messageFromPC
      strtokIndx = strtok(NULL, ",");
      float1FromPC = atof(strtokIndx);     // convert this part to a float
      strtokIndx = strtok(NULL, ",");
      float2FromPC = atof(strtokIndx);     // convert this part to a float

    }

  void apply_settings()
  {
    String txt =  messageFromPC;
    float thresh = float1FromPC;
    if(txt == "ok")
    {
      ok = true;
    }
    else if(txt == "Serial_data_on")
    { 
      serial_data = true;
    }
    else if(txt == "Minimum Inductor current")
    {
     min_IL = thresh;
    }
    else if(txt == "linear_line_on")
    {
      sat_0 = thresh;
      linear_on = true;
      line_on_off = "on";
    }  
    else if(txt == "linear_line_off")
    {
      sat_0 = thresh;
      linear_on = false;
      line_on_off = "off";
    }
    else if(txt == "RL")
    {
      RL = thresh;
    }  
    else if(txt == "curve_fit_on")
    {
      poly_on = true;
      poly_samples = int(thresh);
      cf_on_off = "on";
    }
    else if(txt == "curve_fit_off")
    {
      poly_on = false;
      poly_samples = int(thresh);
      cf_on_off = "off";
    }
    else if(txt == "saturation_limit")
    {
      sat_max = thresh;
    }
    else if(txt == "First_Saturation_detect")
    {
      sat_1 = thresh;
    }
    else if(txt == "Second_Saturation_detect")
    {
      sat_2 = thresh;
    }
    else if(txt == "Third_Saturation_detect")
    {
      sat_3 = thresh;
    }                
      if(thresh > 0.0)
      {
        Serial.print(messageFromPC);
        Serial.print(",");
        if(txt == "RL")
          Serial.println(float1FromPC,3); 
        else if(txt == "curve_fit_on" ||txt == "curve_fit_off")
        {
          Serial.println(float1FromPC,0);    
        }
        else
          Serial.println(float1FromPC);
      }
      else
        Serial.println(messageFromPC);
      
      if(txt == "ok")
      Serial.print('\n');
  }

  void pulse()
  {  
    // Clear all arrays
    for (int i=0; i < 180; i++)
    {
      analog_data[i]=0;// analogRead current data array max pulse_period of 900 / 5us per measurement
      interval[i]=0;// Timer1 count for each inductor current reading starting from 0
      current_data[i]=0.0;// Inductor current in Amps data array
      time_data[i]=0.0;// The time in microseconds for each inductor current reading
      L_data[i]=0.0;// Inductor value in microhenrys data array
      sat_data[i]=0.0;// Inductor saturation in % data array  
    }
    // Reinitialize variables
    I_sat_val_0 = 0;
    I_sat_val_1 = 0;
    I_sat_val_2 = 0;
    I_sat_val_3 = 0;
    t_sat_val_0 = 0;
    t_sat_val_1 = 0;
    t_sat_val_2 = 0;
    t_sat_val_3 = 0;    
    sat_val_0 = 0;
    sat_val_1 = 0;
    sat_val_2 = 0;
    sat_val_3 = 0;
    sat_val_last = 0;
    found_0 = false;
    found_1 = false; 
    found_2 = false; 
    found_3 = false; 
    status_check();
    // 910us = 65535
    // Limit max pulse width to 900us (leave additional room for freq tolerance)
    samples = 0;
    if(pulse_period > 850)
      pulse_period = 850;    
    base_cnt = (pulse_period * 1E-6) / 1.3888E-8;
    gpio_write_bit(GPIOB,9,1);//Output pulse HIGH
    TIMER1_BASE->CNT = 0;  

    // While() loop execution time = 5us
    while(Timer1.getCount() <= base_cnt)
    {   
      interval[samples] = Timer1.getCount();
      analog_data[samples] = analogRead(PA0);// 2.5us    
      samples++;  
    }
    gpio_write_bit(GPIOB,9,0);//Output pulse LOW
    calc();
  }  

  void calc()
  {
    // Convert raw time Timer1 count and analogRead integer data to calculated time(us) and current(Amps) to float values and store in an array
    // Calculate inductance and saturation values and store in array
            
    initial_L = 0;
    sat_val_last = 0; 
    bool initial = false;
      
    // Store the first Time and Inductor current data point
    float time_0 = interval[0] * 1.38888E-2;          // first Time data point in (us)
    float current_0 = (analog_data[0] * .00082) / R_sns;   // first Inductor current data point (Amps)
    
    for(int i=0; i < samples; i++)
    { 
      float time_n = interval[i] * 1.38888E-2;        // Time in (us)
      float current_n = ((analog_data[i] * .00082) / R_sns);  // current (Amps) (50mΩ current sense resistor)
      float delta_time = time_n - time_0;
      float delta_current = current_n - current_0;     
      VL = 12 - ((current_n * R_sns) + (current_n * RL));// Actual inductor voltage minus the Rsense voltage drop and minus the inductors series resistance voltage drop
      //VL = 12.0;// Ideal constant inductor voltage
      
      // For Serial print display ONLY 
      VL_data[i] = VL;

      // Stores time(us) and current(A)
      time_data[i] = time_n;
      current_data[i] = current_n;

      //Calculate and store inductance value
      L_data[i] = VL / (delta_current/delta_time);
      L = L_data[i];
      
      // The initial inductance is captured one time when the initial flag is set
      if(initial)
      {
        if(i == 1)
        {
          initial_L = L_data[i];
          // used for inductance only to display IL on the Oled
          initial_I = current_data[i];
        }
        else
        {
          initial_L = L_data[i-1];
          // used for inductance only to display IL on the Oled
          initial_I = current_data[i-1];
           
        }
        initial = false;
      }
      
      // Set initial flag when the min_IL value is detected in the data sample 
      if(initial_L == 0 && current_n >= min_IL && i < (samples - 1))
      {
        initial = true;       
      }

      //Calculate and store the saturation percentage      
      sat_data[i] = 100 - ((L / initial_L)*100);

      // Calculate the saturation percentage of the last sample
      sat_val_last = 100 - ((L / initial_L)*100);

      // Scale the current and time samples for plotting on the Oled display
      float y_const = 64 / current_data[samples - 1];
      float x_const = 127 / time_data[samples - 1];
      
      // Detects the various user defined saturation limits after initial_L is detected
      if(initial_L != 0 && time_n != initial_t)
      {
        // Detect user selected saturation level for linear line calculation
        if(!found_0 && sat_data[i] >= sat_0 && sat_data[i] < sat_max)
        {
          found_0 = true;
          // Compares the current and the previous saturation values
          // to the user defined limit and selects the closer one
          if(sat_data[i] - sat_0 < sat_0 - sat_data[i - 1])
          {
            I_sat_val_0 = current_data[i];
            t_sat_val_0 = time_data[i];  
            sat_val_0 = sat_data[i];
          }
          else
          {
            I_sat_val_0 = current_data[i - 1];
            t_sat_val_0 = time_data[i - 1];  
            sat_val_0 = sat_data[i - 1];                
          }      
        }
        
        // The first user defined saturation threshold
        if(!found_1 && sat_data[i] >= sat_1 && sat_data[i] < sat_2 && sat_data[i] < sat_3 && sat_data[i] < sat_max)
        {
          found_1 = true;
          if(sat_data[i] - sat_1 < sat_1 - sat_data[i - 1])
          {
            I_sat_val_1 = current_data[i];
            t_sat_val_1 = time_data[i];  
            sat_val_1 = sat_data[i];
          }
          else
          {
            I_sat_val_1 = current_data[i - 1];
            t_sat_val_1 = time_data[i - 1];  
            sat_val_1 = sat_data[i - 1];                
          }
        }
        // The second user defined saturation threshold
        if(!found_2 && sat_data[i] != sat_1 && sat_data[i] >= sat_2 && sat_data[i] < sat_3 && sat_data[i] < sat_max)
        {
          found_2 = true;
          if(sat_data[i] - sat_2 < sat_2 - sat_data[i - 1])
          {
            I_sat_val_2 = current_data[i];
            t_sat_val_2 = time_data[i];  
            sat_val_2 = sat_data[i];
          }
          else
          {
            I_sat_val_2 = current_data[i - 1];
            t_sat_val_2 = time_data[i - 1];  
            sat_val_2 = sat_data[i - 1];                
          }  
        }
        // The third user defined saturation threshold
        if(!found_3 && (sat_data[i] != sat_1 || sat_data[i] != sat_2) && sat_data[i] >= sat_3 && sat_data[i] < sat_max)
        {
          found_3 = true;
          if(sat_data[i] - sat_3 < sat_3 - sat_data[i - 1])
          {
          I_sat_val_3 = current_data[i];
          t_sat_val_3 = time_data[i];  
          sat_val_3 = sat_data[i];
          }
          else
          {
          I_sat_val_3 = current_data[i - 1];
          t_sat_val_3 = time_data[i - 1];  
          sat_val_3 = sat_data[i - 1];                
          }     
        }
    }// End initial_L != 0        
   }// End of samples for loop()
    
    //
    if(inductance_only && initial_L == 0 && pulse_period < 850)
    {
      display.clearDisplay();
      display.display();   
      display.setCursor(36,0);
      display.println("PROCESSING");
      display.setCursor(22,16);
      display.println("INDUCTANCE ONLY");
      display.setCursor(22,32);
      display.println("PW=");      
      display.setCursor(40,32);
      display.println(pulse_period);
      display.setCursor(62,32);
      display.println("us");  
      display.display(); 
      delay(500);  
      pulse_period += 5;
      status_check();             
      pulse();
    }
    //else if(!inductance_only && initial_L == 0)  || (sat_val_last < sat_max && current_data[samples - 1] < 50) && pulse_period < 850)
    else if(!inductance_only && sat_val_last < sat_max && current_data[samples - 1] < 50 && pulse_period < 850)
    {
      display.clearDisplay();
      display.display();   
      display.setCursor(36,0);
      display.println("PROCESSING");
      display.setCursor(22,16);
      display.println("SATURATION TEST");
      display.setCursor(22,32);
      display.println("PW=");      
      display.setCursor(40,32);
      display.println(pulse_period);
      display.setCursor(62,32);
      display.println("us");      
      display.display(); 
      delay(500);
      pulse_period += 5;
      status_check();               
      pulse();  
    }
  }

    void disply()
    {
     if(inductance_only)
     {
      // OLED*************************************************************************
      display.setCursor(2,0);
      display.println(initial_L,1);
      if(initial_L < 10)
      {
        display.setCursor(20,0);
        display.println("uH");
      }
      else if(initial_L >= 0 && initial_L < 100)
      {
        display.setCursor(28,0);
        display.println("uH");
      }
      else if(initial_L >= 00 && initial_L < 1000)
      {
        display.setCursor(32,0);
        display.println("uH");      
      }
      else if(initial_L >= 1000 && initial_L < 10000)
      {
        display.setCursor(38,0);
        display.println("uH"); 
      }
      else
      {
        display.setCursor(42,0);
        display.println("uH"); 
      }
      display.setCursor(2,16);
      display.println("IL=");
      display.setCursor(24,16);
      display.println(initial_I);
      display.setCursor(64,16);
      display.println("Amps");
      display.setCursor(2,32);
      display.println("RL=");
      display.setCursor(24,32);
      display.println(RL,3);
      display.setCursor(64,32);
      display.println("Ohms");
      display.setCursor(2,48);
      display.println("PW=");
      display.setCursor(24,48);
      display.println(pulse_period);
      display.setCursor(46,48);
      display.println("us");      
      display.display();         
     }
     else// Saturation Test
     {
      float y_const = 64 / current_data[samples - 1];
      float x_const = 127 / time_data[samples - 1];

     if(serial_data)
     {
      // Header Text
      Serial.print("t(us)");
      Serial.print('\t');
      Serial.print("  A"); 
      Serial.print('\t');
      Serial.print(" uH");
      Serial.print('\t');
      Serial.print("sat%"); 
      Serial.print('\t');
      Serial.println(" VL");  

      // Display array data
      for(int i=0; i < samples; i++)
      {
        // Serial display
        Serial.print(time_data[i]);
        Serial.print('\t');
        Serial.print(current_data[i]);
        Serial.print('\t');
        Serial.print(L_data[i]);  
        Serial.print('\t');
        Serial.print(sat_data[i]);  
        Serial.print('\t');
        Serial.println(VL_data[i]); 
      }

      // Display Summary  
      Serial.print("Data Count = ");
      Serial.print(samples);
      Serial.println("");
      Serial.println("");
      Serial.println("");
     }                 
      // OLED*************************************************************************
      display.setCursor(2,0);
      display.println(initial_L,1);
      if(initial_L < 10)
      {
        display.setCursor(20,0);
        display.println("uH");
      }
      else if(initial_L >= 0 && initial_L < 100)
      {
        display.setCursor(28,0);
        display.println("uH");
      }
      else if(initial_L >= 00 && initial_L < 1000)
      {
        display.setCursor(32,0);
        display.println("uH");      
      }
      else if(initial_L >= 1000 && initial_L < 10000)
      {
        display.setCursor(38,0);
        display.println("uH"); 
      }
      else
      {
        display.setCursor(42,0);
        display.println("uH"); 
      }
             
      display.setCursor(88,0);
      display.println(current_data[samples - 1],1);
      display.setCursor(114,0);
      display.println("A"); 

      if(current_data[samples - 1] != I_sat_val_1 && current_data[samples - 1] != I_sat_val_2 && current_data[samples - 1] != I_sat_val_3)
      {
        display.setCursor(2,10);
        display.println(sat_data[samples - 1],0);
        display.setCursor(15,10);
        display.println("%max");
      }
        
      display.setCursor(92,56);
      display.println(time_data[samples - 1],0);
      display.setCursor(114,56);
      display.println("us");

      for (int i=0; i <= 64; i++) {
        display.drawPixel(0, i, WHITE);   
      }
      for (int i=0; i < 128; i++) {
        display.drawPixel(i, 63, WHITE);   
      }
      
      // Create 4 tics on the y' axis
      for (int a=0; a < 4; a++){
      for (int i=0; i < 128; i++) {
        if(i >= 126)
        display.drawPixel(i, a * 16, WHITE); 
      }
        
      }
      // Add the first tic current label equal the max current / 4 
      display.setCursor(92,44);
      display.println(current_data[samples - 1]/4,1);
      display.setCursor(120,44);
      display.println("A");

      // Perform a polynomial curve fit of the data if the # of samples is less than 32 
      // If samples are more than 32 then plot the graph using dicrete points only   
      int x;// time time_data
      int y;// current
      if(samples < poly_samples)
      {
        // calculate the polynomial coefficients

        for (int i=0; i <= samples; i++) {
          t_poly[i] = round(x_const * time_data[i]);
          IL_poly[i] = round(y_const * current_data[i]);
        }

        int ret = fitCurve(order, sizeof(IL_poly)/sizeof(double), t_poly, IL_poly, sizeof(coeffs)/sizeof(double), coeffs);
         a = coeffs[0];
         b = coeffs[1];
         c = coeffs[2];
         d = coeffs[3];
         e = coeffs[4]; 
         f = coeffs[5];
         g = coeffs[6];      
      
      // Plot the polynomial of the saturation data     
      for (int i=0; i < 128; i++) {
        x = i;
        if(order == 3)
          y = 64 - (round(a*pow(i,3) + b*pow(i,2) + c*i + d));// 3rd order
        else if(order == 4)
          y = 64 - (round(a*pow(i,4) + b*pow(i,3) + c*pow(i,2) + d*i + e));// 4th order
        else if(order == 5)  
          y = 64 - (round(a*pow(i,5) + b*pow(i,4) + c*pow(i,3) + d*pow(i,2) + e*i + f));// 5th order
        else if(order == 6)
          y = 64 - (round(a*pow(i,6) + b*pow(i,5) + c*pow(i,4) + d*pow(i,3) + e*pow(i,2) + f*i + g));// 6th order
        if(poly_on)
          display.drawPixel(x, y, WHITE);   
      }
      if(poly_on)
      {
        display.setCursor(64,56);
        display.println("CF");
        display.setCursor(77,56);
        display.println(order);              
        display.display();
      }
      }
      if((samples < poly_samples && !poly_on) || samples >= poly_samples)
      {
        // Plot the discrete saturation data points    
        for (int i=0; i < samples; i++) {
          x = round(x_const * time_data[i]);
          y = 64 - (round(y_const * current_data[i]));
          display.drawPixel(x, y, WHITE);   
        }
        display.display();        
      }
              
      // Draw the linear inductance line
      // Normalize and store the first x,y data points 
      float x_1 = round((x_const * t_sat_val_0));//time @ 5% Sat
      float y_1 = round((y_const * I_sat_val_0));//IL @ 5% Sat
      float m = y_1 /x_1;// slope
      float mx = m * x_1;
      for (int i=0; i < 128; i++) {
      int y_ = 64 - round(m*i);

      if(linear_on)
      {
        display.drawPixel(i, y_, WHITE);        
      }
      }     
        if(found_1 && found_2 && found_3)
        {
        // Add 10% value
        display.setCursor(2,40);
        display.println(sat_val_1,0);
        display.setCursor(14,40);
        display.println("%=");
        display.setCursor(26,40);
        display.println(I_sat_val_1,1);         
        display.setCursor(50,40);
        display.println("A");   

        // Add 20% value
        display.setCursor(2,30);
        display.println(sat_val_2,0); 
        display.setCursor(14,30);
        display.println("%=");
        display.setCursor(26,30);
        display.println(I_sat_val_2,1);         
        display.setCursor(50,30);
        display.println("A");  

        // Add 30% Vaue
        display.setCursor(2,20);
        display.println(sat_val_3,0);
        display.setCursor(14,20);
        display.println("%=");
        display.setCursor(26,20);
        display.println(I_sat_val_3,1);         
        display.setCursor(50,20);
        display.println("A");                   
        display.display();

      }
      else if(found_1 && found_2 && !found_3)
      {
        // Add 10% value
        display.setCursor(2,40);
        display.println(sat_val_1,0);
        display.setCursor(14,40);
        display.println("%=");
        display.setCursor(26,40);
        display.println(I_sat_val_1,1);         
        display.setCursor(50,40);
        display.println("A"); 

        // Add 20% value
        display.setCursor(2,30);
        display.println(sat_val_2,0); 
        display.setCursor(14,30);
        display.println("%=");
        display.setCursor(26,30);
        display.println(I_sat_val_2,1);         
        display.setCursor(50,30);
        display.println("A"); 
      }
      else if(found_1 && !found_2 && !found_3)
      {
        // Add 10% value
        display.setCursor(2,40);
        display.println(sat_val_1,0);
        display.setCursor(14,40);
        display.println("%=");
        display.setCursor(26,40);
        display.println(I_sat_val_1,1);         
        display.setCursor(50,40);
        display.println("A"); 
      }
      else if(found_1 && found_3 && !found_2)
      {
        // Add 10% value
        display.setCursor(2,30);
        display.println(sat_val_1,0);
        display.setCursor(14,30);
        display.println("%=");
        display.setCursor(26,30);
        display.println(I_sat_val_1,1);         
        display.setCursor(50,30);
        display.println("A");  

        // Add 30% Vaue
        display.setCursor(2,20);
        display.println(sat_val_3,0);
        display.setCursor(14,20);
        display.println("%=");
        display.setCursor(26,20);
        display.println(I_sat_val_3,1);         
        display.setCursor(50,20);
        display.println("A");                   
        display.display();

      }
      else if(found_2  && !found_1 && !found_3)
      {
        // Add 20% value
        display.setCursor(2,30);
        display.println(sat_val_2,0); 
        display.setCursor(14,30);
        display.println("%=");
        display.setCursor(26,30);
        display.println(I_sat_val_2,1);         
        display.setCursor(50,30);
        display.println("A"); 
      }
        if(found_2 && found_3 && !found_1)
        {  
        // Add 20% value
        display.setCursor(2,30);
        display.println(sat_val_2,0); 
        display.setCursor(14,30);
        display.println("%=");
        display.setCursor(26,30);
        display.println(I_sat_val_2,1);         
        display.setCursor(50,30);
        display.println("A"); 

        // Add 30% Vaue
        display.setCursor(2,20);
        display.println(sat_val_3,0);
        display.setCursor(14,20);
        display.println("%=");
        display.setCursor(26,20);
        display.println(I_sat_val_3,1);         
        display.setCursor(50,20);
        display.println("A");                   
        display.display();
        }
        if(found_3 && !found_1 && !found_2)
        {  
        // Add 30% Vaue
        display.setCursor(2,20);
        display.println(sat_val_3,0);
        display.setCursor(14,20);
        display.println("%=");
        display.setCursor(26,20);
        display.println(I_sat_val_3,1);         
        display.setCursor(50,20);
        display.println("A");                   
        display.display();
        }
      else
      {
        display.display();
      }
    }

      ok = false;
  }
